# -*- coding: utf8 -*-
import click
from terminaltables import PorcelainTable

from mali_commands.commons import print_as_json
from mali_commands.config import Config
from mali_commands.utilities.tables import dict_to_csv


@click.group('auth')
def auth_commands():
    pass


@auth_commands.command('init')
@click.pass_context
@click.option('--webserver/--disable-webserver', default=True, required=False)
def init_auth(ctx, webserver):
    from .commons import pixy_flow

    ctx.obj.local_web_server = webserver

    access_token, refresh_token, id_token = pixy_flow(ctx.obj)

    ctx.obj.config.update_and_save({
        'token': {
            'access_token': access_token,
            'refresh_token': refresh_token,
            'id_token': id_token,
        }
    })


@auth_commands.command('whoami')
@click.pass_context
def whoami(ctx):
    token_data = ctx.obj.config.token_data

    result = {
        'user_id': token_data.get('user_external_id'),
        'name': token_data.get('name'),
        'email': token_data.get('email'),
    }
    json_format = ctx.obj.output_format == 'json'
    format_tables = not json_format

    if format_tables:
        fields = ['name', 'email', 'user_id']
        table_data = list(dict_to_csv(result, fields))

        click.echo(PorcelainTable(table_data).table)
    else:
        print_as_json(result)



# -*- coding: utf8 -*-
from contextlib import closing

import time

from mali_commands.legit import SqliteConnection
from mali_commands.legit.db_index.cache import SqliteTreeCache
from tests.base import BaseTest


class TestTreeCache(BaseTest):
    db_test_path = ':memory:'

    def create_entry(self):
        ctime = time.time()
        mtime = time.time()
        dev = self.some_random_shit_number_int63()
        ino = self.some_random_shit_number_int63()
        mode = self.some_random_shit_number_int63()
        uid = self.some_random_shit_number_int63()
        gid = self.some_random_shit_number_int63()
        gid = self.some_random_shit_number_int63()
        size = self.some_random_shit_number_int63()
        sha = self.some_random_sha_hash()
        flags = self.some_random_shit_number_int63()

        return ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags

    def _test_scan_folder(self):
        entries = {
            'dogs_cats/dogs/0-99/dog_0.jpg': self.create_entry(),
            'dogs_cats/dogs/0-99/dog_1.jpg': self.create_entry(),
        }

        result = list(SqliteTreeCache._folders_to_scan(entries))
        print(result)

    def _test_simple(self):
        with closing(SqliteConnection({}, path=self.db_test_path)) as connection:
            tree_cache = SqliteTreeCache(connection)

            entries = {
                'dogs_cats/dogs/0-99/dog_0.jpg': self.create_entry(),
                'dogs_cats/dogs/0-99/dog_1.jpg': self.create_entry(),
            }

            tree_cache.set_entries(entries)



# -*- coding: utf8 -*-
import requests
from ...backend_mixin import BackendMixin
from .gcs_object_store import GCSObjectStore


class BackendGCSSignedUrlService(BackendMixin):
    def __init__(self, connection, config, handle_api):
        super(BackendGCSSignedUrlService, self).__init__(connection, config, handle_api)

    def get_signed_urls(self, methods, object_names, content_type=None, **kwargs):
        headers = []
        for key, val in kwargs.items():
            headers.append('%s:%s' % (key, val))

        msg = {
            'methods': methods,
            'paths': object_names,
            'headers': headers,
        }
        if 'content-type' in headers:
            del (msg['content_type'])
        if content_type:
            msg['content_type'] = content_type

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=self._volume_id)

        result = self._handle_api(self._config, requests.post, url, msg)
        res = {}
        for method in methods:
            res[method] = result.get(method.lower(), [])
        return res


class BackendGCSObjectStore(BackendMixin, GCSObjectStore):
    def __init__(self, connection, config, handle_api, use_multiprocess=True, processes=-1):
        super(BackendGCSObjectStore, self).__init__(connection, config, handle_api)
        self.__bucket_name = None
        self._signed_url_service = BackendGCSSignedUrlService(connection, config, handle_api)
        self._processes = processes
        self._use_multiprocess = use_multiprocess

    def __iter__(self):
        return super(BackendGCSObjectStore, self).__iter__()

    def close(self):
        super(BackendGCSObjectStore, self).close()

    @classmethod
    def _group_files_by_meta(cls, files):
        content_type_grouped = {}
        for ob in files:
            if ob.content_type not in content_type_grouped:
                content_type_grouped[ob.content_type] = []
            content_type_grouped[ob.content_type].append(ob)
        return content_type_grouped

    def _get_urls_for_paths(self, paths, content_type, headers):
        urls = self._signed_url_service.get_signed_urls(['HEAD', 'PUT'], paths, content_type, **headers)
        head_urls = urls['HEAD']
        put_urls = urls['PUT']
        return head_urls, put_urls

    def _upload_batch(self, content_type, files, callback):
        content_headers = self.get_content_headers()
        upload_paths = list(map(lambda x: GCSObjectStore._get_shafile_path(x.blob.id), files))
        head_urls, put_urls = self._get_urls_for_paths(upload_paths, content_type, content_headers)

        for cur_file, head_url, put_url in zip(files, head_urls, put_urls):
            self.upload(cur_file, content_type, head_url, put_url)
            if callback is not None:
                callback(cur_file)

    def add_objects(self, objects, callback):
        grouped_files = self._group_files_by_meta(objects)
        for content_type in grouped_files:
            files = grouped_files[content_type]
            self._upload_batch(content_type, files, callback)

    @property
    def packs(self):
        return super(BackendGCSObjectStore, self).packs

    def contains_loose(self, sha):
        return super(BackendGCSObjectStore, self).contains_loose(sha)

    def contains_packed(self, sha):
        return super(BackendGCSObjectStore, self).contains_packed(sha)



# -*- coding: utf8 -*-
from tests.base import BaseTest
from mali_commands.utilities.list_utils import flatten


class TestListUtils(BaseTest):

    def testFlatten(self):
        iterable = [[1, 2], [3, 4], [5], (6, 7)]
        expected = [1, 2, 3, 4, 5, 6, 7]

        self.assertEqual(flatten(iterable), expected)



# -*- coding: utf8 -*-
from contextlib import closing
from mali_commands.legit.db import SqliteConnection
from mali_commands.legit.metadata_db import SqliteMetadataDB
from tests.base import BaseTest


class TestMetadata(BaseTest):
    db_test_path = ':memory:'

    def test_add(self):
        sha1 = 'sha1_' + self.some_random_shit()
        sha2 = 'sha2_' + self.some_random_shit()
        tree_id = 'sha2_' + self.some_random_shit()

        with closing(SqliteConnection({}, path=self.db_test_path)) as connection:
            meta = SqliteMetadataDB(connection)

            column_name = 'column_' + self.some_random_shit()

            commits_sha = []

            for i in range(5):
                commits_sha.append('commit_sha%s_%s' % (1, self.some_random_shit()))

            def add_row(index, *args):
                for item_sha in args:
                    val = 2 ** index
                    meta.add_data((item_sha, {'group': {column_name: val}}))

                meta.begin_commit(commits_sha[index], tree_id)
                meta.end_commit()

                return commits_sha[index]

            commits = []
            for i in range(5):
                if (i + 1) % 2 == 0:
                    commits.append(add_row(i, sha1))
                else:
                    commits.append(add_row(i, sha1, sha2))

            self.assertEqual(
                list(meta.get_head_data(sha1)),
                [{'@version': 5, 'group': {column_name: 16}, '@sha': sha1, '@commit_sha': commits_sha[4]}]
            )

            self.assertEqual(
                list(meta.get_head_data(sha2)),
                [{'@version': 5, 'group': {column_name: 16}, '@sha': sha2, '@commit_sha': commits_sha[4]}]
            )

            self.assertEqual(
                meta.get_data_for_commit(sha1, commits[1]),
                {'@version': 2, 'group': {column_name: 2}, '@sha': sha1, '@commit_sha': commits_sha[1]}
            )
            self.assertEqual(
                meta.get_data_for_commit(sha2, commits[1]),
                {'@version': 1, 'group': {column_name: 1}, '@sha': sha2, '@commit_sha': commits_sha[0]}
            )



# -*- coding: utf8 -*-
import datetime
import logging
from ..bigquery_mixin import BigQueryMixin, BqJob
from .base_db_index import BaseMLIndex
from .cache import DatastoreTreeCache, GAEDatastoreTreeCache
from ..db import DatastoreConnection, GAEDatastoreConnection


# noinspection SqlNoDataSourceInspection,SqlResolve
class BigQueryMLIndex(BaseMLIndex, BigQueryMixin):
    STAGING_INDEX_TABLE_NAME = 'staging_index'
    INDEX_TABLE_NAME = 'index'

    def __init__(self, connection, version=None, delete_temp_on_commit=True):
        self.__version = version or 0

        super(BigQueryMLIndex, self).__init__(connection)
        self.__delete_temp_on_commit = delete_temp_on_commit
        self.__create_tree_cache(connection)

    def __create_tree_cache(self, connection):
        if GAEDatastoreTreeCache is not None:
            org_id = self._connection.data_volume_config.org
            volume_id = self._connection.data_volume_config.volume_id

            tree_cache = GAEDatastoreTreeCache(org_id, volume_id)
        else:
            datastore_connection = DatastoreConnection(connection.data_volume_config, connection.project)
            tree_cache = DatastoreTreeCache(datastore_connection)

        self.__tree_cache = tree_cache

    def __get_index_table_ref(self):
        return self._get_table_ref(self.INDEX_TABLE_NAME)

    def __get_staging_index_table_ref(self):
        staging_table_name = '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, self.__version)
        return self._get_table_ref(staging_table_name)

    def __query_count_metadata_and_data_point(self, query):
        from google.cloud.bigquery import QueryJobConfig

        job_config = QueryJobConfig()

        job = self._query_async(query, job_config)

        def process_result(result):
            rows_iter = list(result)
            result = {}
            for is_meta, count in rows_iter:
                field = 'total_meta_data' if is_meta else 'total_data_points'
                result[field] = count

            return result.get('total_data_points', 0), result.get('total_meta_data', 0)

        return BqJob(job, process_result)

    def rpc_version_count_items(self):
        with self._connection.get_cursor() as bq_dataset:
            query = """
                #standardSQL
                SELECT ANY_VALUE(ENDS_WITH(name, '.metadata')), COUNT(DISTINCT name) 
                FROM (SELECT DISTINCT name
                  FROM `{dataset_name}.{index_table_name}`
                  UNION ALL
                    SELECT DISTINCT name
                    FROM `{dataset_name}.{staging_index_table_name}`
                )
                GROUP BY ENDS_WITH(name, '.metadata')
                """.format(
                    dataset_name=bq_dataset.dataset_id,
                    staging_index_table_name=self.__get_staging_index_table_ref().table_id,
                    index_table_name=self.__get_index_table_ref().table_id
            )

            return self.__query_count_metadata_and_data_point(query)

    def rpc_staging_count_items(self):
        staging_table_ref = self.__get_staging_index_table_ref()
        staging_index_table_full_name = staging_table_ref.table_id

        with self._connection.get_cursor() as bq_dataset:
            query = """
                #standardSQL
                SELECT ANY_VALUE(ENDS_WITH(name, '.metadata')), COUNT(DISTINCT name)
                FROM {dataset_name}.{staging_index_table_full_name}
                GROUP BY ENDS_WITH(name, '.metadata')
                """.format(
                    dataset_name=bq_dataset.dataset_id,
                    staging_index_table_full_name=staging_index_table_full_name)

            return self.__query_count_metadata_and_data_point(query)

    @classmethod
    def __table_schema(cls):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('name', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('sha', 'STRING', 'REQUIRED'),
            bigquery.SchemaField('ctime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mtime', 'FLOAT', 'REQUIRED'),
            bigquery.SchemaField('mode', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('size', 'INTEGER', 'REQUIRED'),
            bigquery.SchemaField('url', 'STRING'),
            bigquery.SchemaField('commit_sha', 'STRING', 'REQUIRED'),  # this has to be the last column
            bigquery.SchemaField('ts', 'TIMESTAMP', 'REQUIRED'),  # this has to be the last column
        )

        return schema

    def _create_table_if_needed(self):
        import google.cloud.exceptions
        from google.cloud.bigquery.table import Table

        bq_client = self._connection

        def create_table(method):
            index_table_ref = method()

            table = Table(index_table_ref)
            table.schema = self.__table_schema()

            try:
                bq_client.update_table(table, ['schema'])
            except google.cloud.exceptions.NotFound:
                try:
                    bq_client.create_table(table)
                except google.cloud.exceptions.Conflict:
                    pass

            return table

        create_table(self.__get_index_table_ref)
        create_table(self.__get_staging_index_table_ref)

    def set_entries(self, entries):
        if not entries:
            return

        now = datetime.datetime.utcnow()

        def decode_row_without_gid_uid(entries):
            for name, sha, ctime, mtime, mode, _, _, size, url in self._decode_entries(entries):
                yield name, sha, ctime, mtime, mode, size, url, 'staging', now

        rows = [row for row in decode_row_without_gid_uid(entries)]

        logging.debug('inserting %s rows into bq', len(rows))

        index_table_ref = self.__get_staging_index_table_ref()

        bq_client = self._connection

        bq_client.create_rows(index_table_ref, rows, selected_fields=self.__table_schema())

        self.__tree_cache.set_entries(entries)

        logging.debug('inserted %s rows into bq', len(rows))

    def get_commit_id(self):
        return self.__tree_cache.get_commit_id()

    def __truncate_staging(self):
        if not self.__delete_temp_on_commit:
            logging.debug('index: delete_temp_on_commit: False')
            return

        logging.info('truncate index staging')

        staging_table_ref = self.__get_staging_index_table_ref()
        bq_client = self._connection
        bq_client.delete_table(staging_table_ref)

    def begin_commit(self, commit_sha, tree_id, ts):
        from google.cloud import bigquery

        bq_client = self._connection

        staging_index_table_ref = self.__get_staging_index_table_ref()
        index_table_ref = self.__get_index_table_ref()

        with bq_client.get_cursor() as bq_dataset:
            src_query = """
                #standardSQL
                SELECT staging_index_table.*, @commit_sha as commit_sha, @ts as ts 
                FROM (
                    SELECT * EXCEPT(row_number, ts, commit_sha)
                    FROM (
                      SELECT *, ROW_NUMBER() OVER (PARTITION BY name) row_number
                      FROM `{dataset_name}.{staging_index_table_ref}`
                    )
                    WHERE row_number = 1
                ) staging_index_table
                LEFT JOIN `{dataset_name}.{index_table_name}` index_table
                ON staging_index_table.sha = index_table.sha AND staging_index_table.name = index_table.name
                WHERE index_table.sha is NULL
              """.format(
                dataset_name=bq_dataset.dataset_id,
                staging_index_table_ref=staging_index_table_ref.table_id,
                index_table_name=index_table_ref.table_id,
            )

            src_query_parameters = (
                bigquery.ScalarQueryParameter('commit_sha', 'STRING', commit_sha),
                bigquery.ScalarQueryParameter('ts', 'TIMESTAMP', ts),
            )

            job = self._async_copy_table_data(
                src_query, src_query_parameters, self.__get_index_table_ref())

            return BqJob(job)

    def end_commit(self):
        self.__truncate_staging()

    def delete_all(self):
        self.__tree_cache.delete_all()

        index_table = self._get_table_name(self._connection.table_prefix, self.INDEX_TABLE_NAME)
        staging_index_table_prefix = self._get_table_name(self._connection.table_prefix, self.STAGING_INDEX_TABLE_NAME)
        self._connection.delete_tables([index_table, staging_index_table_prefix])



# -*- coding: utf8 -*-
import six
from .base_metadata_db import BaseMetadataDB, unicode_dict_to_str
from flatten_json import unflatten
from flatten_json import flatten


class SqliteMetadataDB(BaseMetadataDB):
    create_table_sql = """
      CREATE TABLE IF NOT EXISTS metadata (
        '@sha' TEXT NOT NULL,
        '@commit_sha' TEXT NOT NULL,
        '@version' INTEGER,
        PRIMARY KEY ('@sha', '@commit_sha')
      );
      CREATE TABLE IF NOT EXISTS versions (
        'table' TEXT PRIMARY KEY,
        'version' INTEGER NOT NULL
      );
      CREATE INDEX IF NOT EXISTS [@idx_commit_sha] ON [metadata] ([@commit_sha]);
      CREATE INDEX IF NOT EXISTS [@idx_version] ON [metadata] ([@version]);
      INSERT INTO versions([table], [version]) SELECT 'metadata', 0
        WHERE NOT EXISTS(SELECT 1 FROM [versions] WHERE [table] = 'metadata');
    """

    def __init__(self, connection):
        self.__prev_table_info = None

        super(SqliteMetadataDB, self).__init__(connection)

    def __get_table_info(self):
        with self._connection.get_cursor() as c:
            table_info = c.execute("pragma table_info('metadata')").fetchall()

        result = {}
        for column_info in table_info:
            # noinspection PyTypeChecker
            result[column_info['name']] = column_info['type']

        return result

    def _create_table(self):
        with self._connection.get_cursor() as c:
            self._connection.execute_multi(c, self.create_table_sql)

            self._connection.commit()

    @classmethod
    def __value_to_sql_type(cls, column_value):
        if isinstance(column_value, six.string_types):
            return 'TEXT'

        if isinstance(column_value, six.integer_types):
            return 'INTEGER'

        if isinstance(column_value, (float,)):
            return 'REAL'

        raise Exception('UNKNOWN TYPE %s' % type(column_value))

    def __add_column(self, name, column_value):
        column_sql_type = self.__value_to_sql_type(column_value)

        with self._connection.get_cursor() as c:
            c.execute("ALTER TABLE metadata ADD COLUMN '{name}' {type}".format(name=name, type=column_sql_type))
            c.execute("CREATE INDEX 'idx_{name}' ON [metadata] ('{name}')".format(name=name))
            self._connection.commit()

        self.__prev_table_info = self.__get_table_info()

    def __add_missing_columns_from_table_info(self, flatten_data):
        for column_name, column_value in flatten_data.items():
            if column_name not in self.__prev_table_info:
                self.__add_column(column_name, column_value)

    def _add_missing_columns(self, data_object):
        flatten_data = flatten(data_object, separator='.')

        if self.__prev_table_info is None:
            table_info = self.__get_table_info()
            self.__prev_table_info = table_info

        self.__add_missing_columns_from_table_info(flatten_data)

    def _add_data(self, data_list):
        fields = self.__prev_table_info.keys()

        values_params = ['?'] * len(fields)

        values_list = []
        for data_object in data_list:
            flatten_data = flatten(data_object, separator='.')
            values = []

            for field_name in fields:
                values.append(flatten_data.get(field_name))

            values_list.append(values)

        escape_fields = ['[%s]' % field_name for field_name in fields]

        insert_data_sql = "INSERT OR REPLACE INTO [metadata] ({fields}) values ({values_params})".format(
            fields=",".join(escape_fields), values_params=",".join(values_params))

        with self._connection.get_cursor() as c:
            c.executemany(insert_data_sql, values_list)

            self._connection.commit()

    def get_data_for_commit(self, sha, commit_sha):
        get_row_sql = """
            SELECT *
            FROM [metadata]
            WHERE [@sha] = ? AND [@version] IN (
              SELECT MAX([@version])
              FROM [metadata]
              WHERE [@sha] = ? AND [@version] <= (SELECT [@version] FROM [metadata] WHERE [@commit_sha] = ? LIMIT 1)
              GROUP BY [@sha]
            )
        """

        with self._connection.get_cursor() as c:
            c.execute(get_row_sql, (sha, sha, commit_sha,))
            data = c.fetchone()

        if data is None:
            return None

        return unflatten(unicode_dict_to_str(data), separator='.')

    def _query_head_data(self, sha_list):
        format_strings = ','.join(['?'] * len(sha_list))

        get_row_sql = """
            select [metadata].*
            from [metadata]
            inner join (
                SELECT [@sha] as [@__max_sha], MAX([@version]) as [_@max_version]
                FROM [metadata]
                WHERE [@sha] in (%s)
                GROUP BY [@sha]) _max
            on [@sha] = [@__max_sha]
            where [@version] = [_@max_version];
        """ % format_strings

        with self._connection.get_cursor() as c:
            c.execute(get_row_sql, tuple(sha_list))
            for data in c.fetchall():
                yield unflatten(unicode_dict_to_str(data), separator='.')

    def get_all_data(self, sha):
        get_row_sql = """
            SELECT *
            FROM [metadata]
            WHERE [@sha] = ?
            ORDER BY [@version]
        """

        with self._connection.get_cursor() as c:
            c.execute(get_row_sql, (sha,))

            while True:
                results = c.fetchmany(1000)
                if not results:
                    break

                for result in results:
                    yield unflatten(unicode_dict_to_str(result), separator='.')

    def end_commit(self):
        self._connection.commit()

    def begin_commit(self, commit_sha, tree_id):
        update_version_sql = """
          UPDATE [versions] SET [version] = [version] + 1 WHERE [table] = 'metadata';
        """

        update_staging = """
          UPDATE [metadata] SET [@commit_sha] = '{commit_sha}', [@version] = (SELECT [version] FROM [versions] WHERE [table] = 'metadata') WHERE [@commit_sha] = 'staging';
        """

        update_staging = update_staging.format(commit_sha=commit_sha)

        with self._connection.get_cursor() as c:
            c.execute(update_version_sql)
            c.execute(update_staging)

    def _query(self, sql_vars, select_fields, where):
        query_sql = """
             select {select}
             from [metadata]
             where [@version] IN (
               SELECT MAX([@version])
               FROM [metadata]
               GROUP BY [@sha]
             ) AND {where}
             ORDER BY [@sha]
         """.format(where=where, select=','.join(select_fields))

        query_sql = self.fill_in_vars(query_sql, sql_vars)

        for result in self._return_all_result_from_query(query_sql):
            yield result



# -*- coding: utf8 -*-
import os

import six

try:
    import ConfigParser as configparser
except ImportError:
    import configparser


config_file_name = 'config'


def parse_bool(val):
    return val.lower() in ['true', '1']


class DataVolumeConfig(object):
    def __init__(self, path):
        self._config_file = os.path.join(path, config_file_name)

        parser = configparser.RawConfigParser()
        parser.read([self._config_file])

        self.parser = parser
        self._sections = {}

    @property
    def embedded(self):
        val = self.general_config.get('embedded', False)

        if isinstance(val, six.string_types):
            val = parse_bool(val)

        return val

    @property
    def org(self):
        return self.get('org')

    @property
    def volume_id(self):
        return int(self.get('id', mandatory=True))

    def get_boolean(self, section, name, default_value=False, must_exists=False):
        try:
            return self.parser.getboolean(section, name)
        except (configparser.NoSectionError, configparser.NoOptionError):
            if must_exists:
                raise

            return default_value

    @property
    def is_local(self):
        return self.get_boolean('general', 'is_local')

    @property
    def db_type(self):
        return self.db_config.get('type')

    @property
    def object_store_type(self):
        return self.object_store_config.get('type')

    def get(self, key, mandatory=False):
        if mandatory:
            return self.general_config[key]

        return self.general_config.get(key)

    @property
    def data_path(self):
        return self.general_config.get('datapath')

    @property
    def general_config(self):
        return self.items('general')

    @property
    def db_config(self):
        return self.items('db')

    @property
    def object_store_config(self):
        return self.items('object_store')

    def items_from_parse(self, parser, section, must_exists):
        try:
            if section not in self._sections:
                self._sections[section] = dict(parser.items(section))

            return self._sections[section]
        except configparser.NoSectionError:
            if must_exists:
                raise

            return self._sections.setdefault(section, {})

    def items(self, section, most_exists=False):
        return self.items_from_parse(self.parser, section, must_exists=most_exists)

    def set(self, section, key, val):
        try:
            self.parser.add_section(section)
        except configparser.DuplicateSectionError:
            pass

        self.parser.set(section, key, val)

    def _write(self, fo):
        self.parser.write(fo)

    def save(self):
        with open(self._config_file, 'w') as configfile:
            self._write(configfile)

    def update_and_save(self, d):
        for section, section_values in d.items():
            for key, val in section_values.items():
                if val is None:
                    continue

                self.set(section, key, val)

        self.save()



# -*- coding: utf8 -*-
import requests
import fudge
import json

from fudge.inspector import arg

from click.testing import CliRunner

from tests.base import BaseCliTest
from mali import cli
from mali_commands.experiments import chart_name_to_id


class TestExperiments(BaseCliTest):
    project_id = BaseCliTest.some_random_shit_number_int63()
    experiment_id = BaseCliTest.some_random_shit_number_int63()
    user_sent_metrics = {
        BaseCliTest.some_random_shit(): BaseCliTest.some_random_shit()
    }

    @fudge.patch('mali_commands.commons.handle_api')
    def testListExperiments(self, handle_api_mock):
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.get,
            'projects/{project_id}/experiments'.format(project_id=self.project_id)
        ).returns({
            'experiments': [
                {
                    'experiment_id': BaseCliTest.some_random_shit_number_int63(),
                    'created_at': '2017-05-04T07:18:47.620155',
                    'display_name': BaseCliTest.some_random_shit(),
                    'description': BaseCliTest.some_random_shit()
                }
            ]
        })

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'list', '--projectId', self.project_id], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def testUpdateMetrics_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.post,
            'projects/{project_id}/experiments/{experiment_id}/metrics'.format(project_id=self.project_id,
                                                                               experiment_id=self.experiment_id),
            self.user_sent_metrics
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateMetrics', '--projectId', self.project_id,
                                     '--experimentId', self.experiment_id, '--metrics',
                                     json.dumps(self.user_sent_metrics)], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def testGraphUpdateUpdateGraph_withValidJSONAndShortHands_callHandleApiAndExit(self, handle_api_mock):
        import json

        chart_name = self.some_random_shit(size=10)
        chart_id = chart_name_to_id(chart_name)

        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = self.some_random_shit(size=10)
        chart_size = 10
        chart_x = list(range(0, chart_size))
        chart_y = list(range(0, chart_size))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name,
                'x_int': chart_x, 'y_data': chart_y, 'chart': chart_type, 'scope': scope, 'labels': [chart_x_name, chart_y_name]}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.put,
            'projects/{project_id}/experiments/{experiment_id}/chart/{chart_id}'.format(project_id=self.project_id,
                                                                                        experiment_id=self.experiment_id,
                                                                                        chart_id=chart_id),
            data
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart', '-p', self.project_id,
                                     '-e', self.experiment_id, '-c', chart_name,
                                     '-cs', scope, '-ct', chart_type,
                                     '-cxn', chart_x_name, '-cyn', chart_y_name,
                                     '-cx', json.dumps(chart_x), '-cy', json.dumps(chart_y)],
                               catch_exceptions=False
                               )
        self.assertEqual(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def testGraphUpdateUpdateGraph_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        import json

        chart_name = self.some_random_shit(size=10)
        chart_id = chart_name_to_id(chart_name)

        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = self.some_random_shit(size=10)
        chart_size = 10
        chart_x = list(range(0, chart_size))
        chart_y = list(range(0, chart_size))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name,
                'x_int': chart_x, 'y_data': chart_y, 'chart': chart_type, 'scope': scope, 'labels': [chart_x_name, chart_y_name]}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.put,
            'projects/{project_id}/experiments/{experiment_id}/chart/{chart_id}'.format(project_id=self.project_id,
                                                                                        experiment_id=self.experiment_id,
                                                                                        chart_id=chart_id),
            data
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart', '--projectId', self.project_id,
                                     '--experimentId', self.experiment_id, '--chartName', chart_name,
                                     '--chartScope', scope, '--chartType', chart_type,
                                     '--chartXName', chart_x_name, '--chartYName', chart_y_name,
                                     '--chartX', json.dumps(chart_x), '--chartY', json.dumps(chart_y)],
                               catch_exceptions=False
                               )
        self.assertEqual(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def testGraphUpdateUpdateGraph_withSizeMismach_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        import json

        chart_id = self.some_random_shit(size=10)
        chart_name = self.some_random_shit(size=10)
        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = self.some_random_shit(size=10)
        chart_size = 10
        chart_x = list(range(0, chart_size + 1))
        chart_y = list(range(chart_size, 2 * chart_size))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name, 'x_name': chart_x_name, 'y_name': chart_y_name,
                'x': chart_x, 'y': chart_y, 'chart': chart_type,
                'scope': scope}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart', '--projectId', self.project_id,
                                     '--experimentId', self.experiment_id, '--chartName', chart_id,
                                     '--scope', scope, '--chartType', chart_type, '--chartName', chart_name,
                                     '--chartXName', chart_x_name, '--chartYName', chart_y_name,
                                     '--chartX', json.dumps(chart_x), '--chartY', json.dumps(chart_y)],
                               catch_exceptions=True)
        self.assertEqual(result.exit_code, 2)

    @fudge.patch('mali_commands.commons.handle_api')
    def testGraphUpdateUpdateGraph_with_no_proj_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        import json

        chart_id = self.some_random_shit(size=10)
        chart_name = self.some_random_shit(size=10)
        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = self.some_random_shit(size=10)
        chart_size = 10
        chart_x = list(range(0, chart_size))
        chart_y = list(range(chart_size, 2 * chart_size))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name, 'x_name': chart_x_name, 'y_name': chart_y_name,
                'x': chart_x, 'y': chart_y, 'chart': chart_type,
                'scope': scope}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart',
                                     '--experimentId', self.experiment_id, '--chartName', chart_id,
                                     '--chartScope', scope, '--chartType', chart_type, '--chartName', chart_name,
                                     '--chartXName', chart_x_name, '--chartYName', chart_y_name,
                                     '--chartX', json.dumps(chart_x), '--chartY', json.dumps(chart_y)],
                               catch_exceptions=True)
        self.assertEqual(result.exit_code, 2)

    def testGraphUpdateUpdateGraph_with_bad_y_size_withValidJSON_callHandleApiAndExit(self):
        import json

        chart_id = self.some_random_shit(size=10)
        chart_name = self.some_random_shit(size=10)
        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = [self.some_random_shit(size=10), self.some_random_shit(size=10)]
        chart_size = 10
        chart_x = list(range(0, chart_size))
        chart_y = list(range(2 * chart_size + 1))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name, 'x_name': chart_x_name, 'y_name': chart_y_name,
                'x': chart_x, 'y': chart_y, 'chart': chart_type,
                'scope': scope}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart', '--projectId', self.project_id,
                                     '--experimentId', self.experiment_id, '--chartName', chart_id,
                                     '--chartScope', scope, '--chartType', chart_type, '--chartName', chart_name,
                                     '--chartXName', chart_x_name, '--chartYName', json.dumps(chart_y_name),
                                     '--chartX', json.dumps(chart_x), '--chartY', json.dumps(chart_y)],
                               catch_exceptions=True)
        self.assertEqual(result.exit_code, 2)
        self.assertEqual(result.output, 'Usage: cli experiments updateChart [OPTIONS]\n\nError: X and Y arrays must be of the same size\n')

    @fudge.patch('mali_commands.commons.handle_api')
    def testGraphUpdateUpdateGraph_with_no_experiment_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        import json

        chart_id = self.some_random_shit(size=10)
        chart_name = self.some_random_shit(size=10)
        chart_x_name = self.some_random_shit(size=10)
        chart_y_name = self.some_random_shit(size=10)
        chart_size = 11
        chart_x = list(range(0, chart_size))
        chart_y = list(range(chart_size, 2 * chart_size))
        chart_type = 'line'
        scope = 'test'
        data = {'name': chart_name, 'x_name': chart_x_name, 'y_name': chart_y_name,
                'x': chart_x, 'y': chart_y, 'chart': chart_type,
                'scope': scope}
        request = dict(data)
        request['project_id'] = self.project_id
        request['experiment_id'] = self.experiment_id
        request['chart_id'] = chart_id

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateChart', '--projectId', self.project_id,
                                     '--scope', scope, '--chartType', chart_type, '--chartName', chart_name,
                                     '--chartXName', chart_x_name, '--chartYName', chart_y_name,
                                     '--chartX', json.dumps(chart_x), '--chartY', json.dumps(chart_y)],
                               catch_exceptions=True)
        self.assertEqual(result.exit_code, 2)

    @fudge.patch('mali_commands.commons.handle_api')
    def testUpdateMetrics_withValidJSONAndShortHands_callHandleApiAndExit(self, handle_api_mock):
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.post,
            'projects/{project_id}/experiments/{experiment_id}/metrics'.format(project_id=self.project_id,
                                                                               experiment_id=self.experiment_id),
            self.user_sent_metrics
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateMetrics', '-p', self.project_id,
                                     '-e', self.experiment_id, '-m', json.dumps(self.user_sent_metrics)],
                               catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

    @fudge.patch('mali_commands.commons.handle_api')
    def testUpdateMetrics_withWeightsHash_callHandleApiAndExit(self, handle_api_mock):
        model_weights_hash = BaseCliTest.some_random_shit()
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.post,
            'model_weights_hashes/{model_weights_hash}/metrics?experiment_only=1'.format(
                model_weights_hash=model_weights_hash),
            self.user_sent_metrics
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateMetrics', '--weightsHash', model_weights_hash,
                                     '--metrics', json.dumps(self.user_sent_metrics)],
                               catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

    def testUpdateMetrics_projectIdSpecifiedWithoutExperimentId_raiseBadOptionUsage(self):
        runner = CliRunner()
        params = ['experiments', 'updateMetrics', '--projectId', self.project_id,
                  '--metrics', BaseCliTest.some_random_shit()]
        result = runner.invoke(cli, params, catch_exceptions=False)
        self.assertEqual(result.output, 'Usage: cli experiments updateMetrics [OPTIONS]\n\n'
                                        'Error: Please also specify --experimentId option.\n')

    def testUpdateMetrics_experimentIdSpecifiedWithoutProjectId_raiseBadOptionUsage(self):
        runner = CliRunner()
        params = ['experiments', 'updateMetrics', '--experimentId', self.experiment_id,
                  '--metrics', BaseCliTest.some_random_shit()]
        result = runner.invoke(cli, params, catch_exceptions=False)
        self.assertEqual(result.output, 'Usage: cli experiments updateMetrics [OPTIONS]\n\n'
                                        'Error: Please also specify --projectId option.\n')

    def testUpdateMetrics_noOptionSpecified_raiseBadOptionUsage(self):
        runner = CliRunner()
        params = ['experiments', 'updateMetrics', '--metrics', BaseCliTest.some_random_shit()]
        result = runner.invoke(cli, params, catch_exceptions=False)
        self.assertEqual(result.output, 'Usage: cli experiments updateMetrics [OPTIONS]\n\n'
                                        'Error: Please specify the experiment using (1) --projectId and '
                                        '--experimentId optionsor (2) --weightsHash options.\n')

    def testUpdateMetrics_withInvalidJSON_raiseValueError(self):
        runner = CliRunner()
        params = ['experiments', 'updateMetrics', '--projectId', self.project_id, '--experimentId', self.experiment_id,
                  '--metrics', BaseCliTest.some_random_shit()]
        result = runner.invoke(cli, params, catch_exceptions=False)
        self.assertEqual(result.output, 'Usage: cli experiments updateMetrics [OPTIONS]\n\n'
                                        'Error: Invalid value: The supplied sting is not a valid JSON dictionary format.\n')

    @fudge.patch('mali_commands.commons.handle_api')
    def testUpdateModelMetrics_withValidJSON_callHandleApiAndExit(self, handle_api_mock):
        model_weights_hash = BaseCliTest.some_random_shit()
        handle_api_mock.expects_call().with_matching_args(
            arg.any(),  # ctx
            requests.post,
            'model_weights_hashes/{model_weights_hash}/metrics'.format(model_weights_hash=model_weights_hash),
            self.user_sent_metrics
        ).returns({})

        runner = CliRunner()
        result = runner.invoke(cli, ['experiments', 'updateModelMetrics', '--weightsHash', model_weights_hash,
                                     '--metrics', json.dumps(self.user_sent_metrics)], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)



# -*- coding: utf8 -*-
from .mysql import MySqlConnection
from .sqlite import SqliteConnection
from .spanner import SpannerConnection
from .bq import BigQueryConnection, BigQuerySqlHelper
from .datastore import DatastoreConnection
from .gae_datastore import GAEDatastoreConnection
from .backend_conection import BackendConnection



# -*- coding: utf8 -*-
from .base_connection import BaseConnection


class SpannerSqlHelper(object):
    def __init__(self):
        pass

    @classmethod
    def escape(cls, name):
        return '`%s`' % name

    @classmethod
    def random_function_name(cls):
        return 'RAND'


class SpannerConnection(BaseConnection):
    def __init__(self, data_volume_config, project, instance, database, **kwargs):
        self.__database = database
        self.__instance = instance
        self.__project = project
        super(SpannerConnection, self).__init__(data_volume_config, **kwargs)

    def _create_connection(self, **kwargs):
        from google.cloud import spanner

        spanner_client = spanner.Client(project=self.__project, credentials=credentials)
        return spanner_client.instance(self.__instance)

    def _create_cursor(self):
        return self._native_conn.database(self.__database)

    def _commit(self):
        pass

    def _rollback(self):
        pass

    def create_sql_helper(self):
        return SpannerSqlHelper()



# # -*- coding: utf8 -*-
# import fudge
# import httpretty
# from fudge.inspector import arg
#
# from mali import cli
# from .base import BaseCliTest
# from click.testing import CliRunner
#
#
# class TestAuth(BaseCliTest):
#     user_org = "user_org_" + BaseCliTest.some_random_shit()
#     name = "name_" + BaseCliTest.some_random_shit()
#     git_endpoint = "git_endpoint_" + BaseCliTest.some_random_shit()
#
#     @httpretty.activate
#     @fudge.patch('mali_commands.commons.handle_api')
#     def test_create_runcode(self, handle_api):
#         handle_api.expects_call().with_matching_args(
#             arg.any(),  # ctx
#             arg.any(),  # http method
#             'runcode',
#             {
#                 'name': self.name,
#                 'git': self.git_endpoint,
#                 'org': self.user_org,
#             }
#         ).returns({})
#
#         runner = CliRunner()
#         result = runner.invoke(
#             cli,
#             ["code", "create", "--org", self.user_org, "--name", self.name, "--git", self.git_endpoint],
#             catch_exceptions=False)
#         self.assertEquals(result.exit_code, 0)



# -*- coding: utf8 -*-
import jinja2
import os
from flask import Flask, render_template, request, send_from_directory

from mali_commands.data_volume import with_repo

app = Flask(__name__)

template_folder = os.path.join(os.path.dirname(__file__), 'templates')

my_loader = jinja2.ChoiceLoader([
    app.jinja_loader,
    jinja2.FileSystemLoader([template_folder]),
])

app.jinja_loader = my_loader


@app.route('/images/<path:path>')
def create_images_handler(path):
    data_path = app.config['data_path']

    return send_from_directory(data_path, path)


@app.route("/")
def main():
    query = request.args.get('q')

    repo_root = app.config['repo_root']
    config_prefix = app.config['config_prefix']

    with with_repo(config_prefix, repo_root) as repo:
        return render_template('index.html', images=repo.metadata.query(query))



# -*- coding: utf8 -*-
import logging

try:
    from .backend_gcs_object_store import *
except ImportError as ex:
    logging.debug("ImportError: %s", ex)
    BackendGCSObjectStore = None

try:
    from .gcs_object_store import GCSDeleteAll, GCSObjectStore
except ImportError as ex:
    logging.debug("ImportError: %s", ex)
    GCSObjectStore = None
    GCSDeleteAll = None



# -*- coding: utf8 -*-
import logging
import datetime
from collections import OrderedDict

import requests

from ...connection_mixin import ConnectionMixin
from ...dulwich.object_store import BaseObjectStore
from ...dulwich.objects import hex_to_filename, sha_to_hex, Blob

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class GCSServiceOperationError(Exception):
    pass


__gcs_service = None
__gcs_credentials = None


def gcs_credentials():
    import google.auth.exceptions

    global __gcs_credentials

    if __gcs_credentials is None:
        try:
            __gcs_credentials = google.auth.default()
        except google.auth.exceptions.DefaultCredentialsError:
            pass

    return __gcs_credentials


def gcs_service(credentials=None):
    from google.cloud import storage

    global __gcs_service

    if __gcs_service is None:
        __gcs_service = storage.Client(credentials=credentials)

    return __gcs_service


class GCSService(object):
    RETRYABLE_ERRORS = (IOError, )
    DEFAULT_MIMETYPE = 'application/octet-stream'
    NUM_RETRIES = 5

    def __init__(self, signed_url_service):
        self.__service = None
        self.signed_url_service = signed_url_service


def do_upload(bucket_name, volume_id, object_name, full_path_to_data, signed_url_service, content_type, head_url, put_url, credentials=None):
    GCSUpload(signed_url_service).upload(
        bucket_name, volume_id, object_name, full_path_to_data, content_type, head_url, put_url, credentials=credentials)


def do_download(bucket_name, volume_id, object_name, signed_url_service=None):
    return GCSDownload(signed_url_service).download(bucket_name, volume_id, object_name)


def do_delete_all(bucket_name, volume_id, max_files):
    return GCSDeleteAll(signed_url_service=None).delete_all(bucket_name, volume_id, max_files)


def __raise_on_request_error(http_operation):

    return http_operation()


def _retry_operation(http_operation, on_failed):
    start_time = datetime.datetime.utcnow()

    retries = 0
    done = False
    while not done:
        try:
            progress, done = __raise_on_request_error(http_operation)
        except GCSService.RETRYABLE_ERRORS as ex:
            logging.debug('retry operation because of %s', ex)
            retries += 1

            if retries == GCSService.NUM_RETRIES:
                on_failed()
                raise GCSServiceOperationError()

            continue

    logger.debug('operation took %s', datetime.datetime.utcnow() - start_time)


def return_request_result(response, content):
    return response, content


class GCSDeleteAll(GCSService):
    @classmethod
    def delete_all(cls, bucket_name, volume_id, max_files=None):
        import google.cloud.exceptions

        logging.info('delete all at %s/%s', bucket_name, volume_id)
        gcs = gcs_service()

        try:
            list_iter = gcs.bucket(bucket_name).list_blobs(prefix=volume_id)
        except google.cloud.exceptions.NotFound:
            logging.warning('bucket %s was not found', bucket_name)
            return

        total_deleted = 0
        for blob in list_iter:
            try:
                gcs.bucket(bucket_name).delete_blob(blob.name)
            except google.cloud.exceptions.NotFound:
                pass

            total_deleted += 1

            if max_files is not None and max_files == total_deleted:
                break

        logging.info('total deleted %s', total_deleted)

        return total_deleted


class GCSDownload(GCSService):
    def download(self, bucket_name, volume_id, object_name):
        if not self.signed_url_service:
            gcs = gcs_service()

            object_name = '%s/%s' % (volume_id, object_name)
            blob = gcs.bucket(bucket_name).blob(object_name)
            data = blob.download_as_string()
        else:
            signed_urls = self.signed_url_service.get_signed_urls(['GET'], [object_name])
            url = signed_urls['GET'][0]

            r = requests.get(url)
            r.raise_for_status()
            data = r.content

        logger.debug('downloaded  %s(%s)', object_name, len(data))

        return data


class GCSUpload(GCSService):
    def upload(self, bucket_name, volume_id, object_name, full_path_to_data, content_type=None, head_url=None, put_url=None, credentials=None):
        from mali_commands.utilities import get_content_type

        with open(full_path_to_data, 'rb') as f:
            body = f.read()

        logger.debug('upload %s/%s (%s)', volume_id, object_name, '{:,}'.format(len(body)))

        def validate_urls(head, put, current_content_type):
            if head is None or put is None:
                if current_content_type is None:
                    current_content_type = get_content_type(body)

                urls = self.signed_url_service.get_signed_urls(
                    ['PUT', 'HEAD'], [object_name], current_content_type, **GCSObjectStore.get_content_headers())
                put = urls['PUT'][0]
                head = urls['HEAD'][0]

            return current_content_type, head, put

        def do_cold_upload():
            if self.signed_url_service or put_url:
                return False

            current_gcs = gcs_service(credentials=credentials)

            direct_object_name = '%s/%s' % (volume_id, object_name)

            blob = current_gcs.bucket(bucket_name).blob(direct_object_name)
            if blob.exists():
                return True

            blob.upload_from_string(body, get_content_type(body))

            return True

        upload_res = do_cold_upload()
        if upload_res:
            return

        content_type, head_url, put_url = validate_urls(head_url, put_url, content_type)

        resp = requests.head(head_url)

        if resp.status_code in (204, 404):
            c_type = content_type or get_content_type(body)
            c_headers = GCSObjectStore.get_content_headers(c_type)

            logging.debug('file not found, uploading')

            r = requests.put(put_url, data=body, headers=c_headers)
            r.raise_for_status()


class GCSObjectStore(ConnectionMixin, BaseObjectStore):
    def __init__(self, connection, use_multiprocess=True, processes=-1):
        super(GCSObjectStore, self).__init__(connection)
        self.__upload_pool = None
        self._use_multiprocess = use_multiprocess
        self._multi_process_control = None
        self._processes = processes
        self.__bucket_name = connection.data_volume_config.object_store_config.get('bucket_name')
        self.__volume_id = self._connection.data_volume_config.volume_id
        self._signed_url_service = None

    def delete_all(self, max_files=None):
        return do_delete_all(self.__bucket_name, self.__volume_id, max_files)

    @classmethod
    def get_content_headers(cls, content_type=None):

        headers = OrderedDict()
        if content_type:
            headers['Content-Type'] = content_type

        headers['x-goog-acl'] = 'public-read'
        headers['x-goog-if-generation-match'] = '0'

        return headers

    @property
    def processes(self):
        return self._processes

    @processes.setter
    def processes(self, value):
        self._processes = value

    @property
    def is_multiprocess(self):
        return self._use_multiprocess and self._processes != 1

    def close(self):
        logging.debug('%s closing', self.__class__)
        if self._multi_process_control is not None:
            self._multi_process_control.close()

        logging.debug('%s closed', self.__class__)

    @classmethod
    def _get_shafile_path(cls, sha):
        # Check from object dir
        return hex_to_filename('objects', sha)

    @classmethod
    def on_upload_error(cls, ex):
        raise ex

    def upload(self, obj, content_type=None, head_url=None, put_url=None, callback=None):
        path = self._get_shafile_path(obj.blob.id)

        credentials = gcs_credentials()

        if self.is_multiprocess:
            if self._multi_process_control is None:
                from ...multi_process_control import get_multi_process_control

                self._multi_process_control = get_multi_process_control(self._processes)

            self._multi_process_control.execute(
                do_upload,
                args=(self.__bucket_name, self.__volume_id, path, obj.full_path, None, content_type, head_url, put_url, credentials), callback=callback)

        else:
            do_upload(self.__bucket_name, self.__volume_id, path, obj.full_path, self._signed_url_service, content_type, head_url, put_url)
            if callback is not None:
                callback(obj)

    def add_object(self, obj):

        """Add a single object to this object store.

        :param obj: Object to add
        """
        self.upload(obj)

    def _get_loose_object(self, sha):
        logger.debug('get object %s', sha)
        path = self._get_shafile_path(sha)
        data = do_download(self.__bucket_name, self.__volume_id, path, signed_url_service=self._signed_url_service)
        blob = Blob()
        blob.set_raw_chunks([data], sha)
        return blob

    def get_raw(self, name):
        """Obtain the raw text for an object.

        :param name: sha for the object.
        :return: tuple with numeric type and object contents.
        """
        hex_sha = name

        if len(name) != 40 and len(name) != 20:
            raise AssertionError("Invalid object name %r" % name)

        if hex_sha is None:
            hex_sha = sha_to_hex(name)

        ret = self._get_loose_object(hex_sha)
        if ret is not None:
            return ret.type_num, ret.as_raw_string()

        raise KeyError(hex_sha)

    @property
    def packs(self):
        raise NotImplementedError(self.packs)

    def __iter__(self):
        raise NotImplementedError(self.__iter__)

    def add_objects(self, objects, callback=None):
        for obj in objects:
            self.upload(obj, callback=callback)

    def contains_packed(self, sha):
        raise NotImplementedError(self.contains_packed)

    def contains_loose(self, sha):
        raise NotImplementedError(self.contains_loose)



# import click
# import requests
# from .commons import add_to_data_if_not_none, output_result
#
#
# @click.group('code')
# def runcode_commands():
#     pass
#
#
# @runcode_commands.command('create')
# @click.pass_context
# @click.option('--name', required=True)
# @click.option('--git', required=True)
# @click.option('--org', required=False)
# def code_create(ctx, name, git, org):
#     data = {}
#
#     add_to_data_if_not_none(data, name, "name")
#     add_to_data_if_not_none(data, git, "git")
#     add_to_data_if_not_none(data, org, "org")
#
#     result = ctx.obj.handle_api(ctx.obj, requests.post, 'runcode', data)
#
#     output_result(ctx, result, ['ok'])



# -*- coding: utf8 -*-
from click.testing import CliRunner

from tests.base import BaseCliTest
from mali import cli
from mali_commands.models import get_keras_weights_hash, get_tensorflow_weights_hash


class TestModelsWithKeras(BaseCliTest):
    def testGetWeightsHash(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['models', 'getWeightsHash',
                                     '--framework', 'keras',
                                     '--model', 'tests/assets/keras_full_model.hdf5'], catch_exceptions=False)

        self.assertEqual(result.exit_code, 0)

    def testGetKerasWeightsHash(self):
        model_filepath = 'tests/assets/keras_full_model.hdf5'
        weights_hash = get_keras_weights_hash(model_filepath)
        self.assertEqual(weights_hash, 'v1_695a3231a3c842902f3f47c1e125c2b259b4bc89')



# -*- coding: utf8 -*-
from .sqlite_db_index import SqliteMLIndex
from .mysql_db_index import MySqlMLIndex
from .spanner_db_index import SpannerMLIndex
from .bigquery_db_index import BigQueryMLIndex
from .backend_db_index import BackendMLIndex



# -*- coding: utf8 -*-
from .base_db_index import BaseMLIndex


class SpannerMLIndex(BaseMLIndex):
    create_index_ddl = """
      CREATE TABLE files (
        name String(256) NOT NULL,
        sha String(256) NOT NULL,
        ctime INT64	NOT NULL,
        mtime INT64	NOT NULL,
        mode INT64 NOT NULL,
        uid INT64 NOT NULL,
        gid INT64 NOT NULL,
        size INT64 NOT NULL
      ) PRIMARY KEY (name)
    """

    def set_entries(self, entries):
        values = self._decode_entries(entries)

        with self._connection.get_cursor() as database:
            with database.batch() as batch:
                batch.insert(
                    table='files',
                    columns=('name', 'sha', 'ctime', 'mtime', 'mode', 'uid', 'gid', 'size', ),
                    values=values)

    def _create_table_if_needed(self):
        with self._connection.get_cursor() as c:
            c.update_ddl([self.create_index_ddl])

    def __init__(self, connection):
        super(SpannerMLIndex, self).__init__(connection)



# -*- coding: utf8 -*-
import mimetypes

mimetypes.init()
mimetypes.add_type(mimetypes.types_map.get('.jpg'), '.jfif')


def get_content_type(body):
    import puremagic

    try:
        ext = puremagic.from_string(body)
        return mimetypes.types_map.get(ext)
    except puremagic.PureError:
        return None



# -*- coding: utf8 -*-
from .base_db_index import BaseMLIndex


class MySqlMLIndex(BaseMLIndex):
    create_index_sql = """
      CREATE TABLE
      IF NOT EXISTS files (
        name varchar(256) PRIMARY KEY,
        sha text not null,
        ctime integer not null,
        mtime integer not null,
        mode integer not null,
        uid integer not null,
        gid integer not null,
        size integer not null
      );
    """

    def __init__(self, connection):
        super(MySqlMLIndex, self).__init__(connection)

    def remove_path(self, name):
        with self._connection.get_cursor() as c:
            c.execute("DELETE from files where name=%s limit 1", (name.decode('utf8'), ))
            self._connection.commit()

    def set_entries(self, entries):
        values = self._decode_entries(entries)

        with self._connection.get_cursor() as c:
            c.executemany("REPLACE INTO files (name, sha, ctime, mtime, mode, uid, gid, size) VALUES (%s, %s, %s, %s, %s, %s, %s, %s)", values)
            self._connection.commit()

    def _create_table_if_needed(self):
        with self._connection.get_cursor() as c:
            c.execute(self.create_index_sql)
            self._connection.commit()

    def iterblobs(self):
        with self._connection.get_cursor() as c:
            c.execute('SELECT name, sha, mode FROM files order by name')
            for row in c:
                yield row[0].encode('utf8'), row[1].encode('utf8'), row[2]

    def commit(self, object_store):
        from dulwich.index import commit_tree

        return commit_tree(object_store, self.iterblobs())
